import httpx


class EapiCommandError(RuntimeError):
    """
    Exception class for EAPI command errors

    Attributes
    ----------
    failed: str - the failed command
    errmsg: str - a description of the failure reason
    passed: List[dict] - a list of command results of the commands that passed
    not_exec: List[str] - a list of commands that were not executed
    """

    def __init__(self, failed, errmsg, passed, not_exec):
        """Initializer for the EapiCommandError exception"""
        self.failed = failed
        self.errmsg = errmsg
        self.passed = passed
        self.not_exec = not_exec
        super(EapiCommandError, self).__init__()


# alias for exception during sending-receiving
EapiTransportError = httpx.HTTPStatusError



from typing import Optional, AnyStr, Tuple, List

from socket import getservbyname
import httpx
import base64
from collections import namedtuple


__all__ = ["Device", "CommandResults"]


CommandResults = namedtuple("CommandResults", ["ok", "command", "output"])


class Transport(object):
    OFMT_OPTIONS = ("json", "text")

    def __init__(self, host, proto, port, creds, timeout=60):
        port = port or getservbyname(proto)
        self.client = httpx.AsyncClient(
            base_url=httpx.URL(f"{proto}://{host}:{port}"),
            verify=False,
            timeout=httpx.Timeout(timeout),
        )

        self.client.headers["Content-Type"] = "application/json-rpc"
        self.b64auth = (
            base64.encodebytes(bytes("%s:%s" % creds, encoding="utf-8"))
            .decode()
            .replace("\n", "")
        )
        self.client.headers["Authorization"] = "Basic %s" % self.b64auth

        self.username = creds[0]
        self.ofmt = "json"

    @property
    def timeout(self):
        return self.client.timeout

    @timeout.setter
    def timeout(self, value):
        self.client.timeout = httpx.Timeout(value)

    def form_command(self, commands, **kwargs) -> dict:
        cmd = {
            "jsonrpc": "2.0",
            "method": "runCmds",
            "params": {
                "version": 1,
                "cmds": commands,
                "format": kwargs.get("ofmt") or self.ofmt,
            },
            "id": str(kwargs.get("req_id") or id(self)),
        }
        if "autoComplete" in kwargs:
            cmd["params"]["autoComplete"] = kwargs["autoComplete"]

        if "expandAliases" in kwargs:
            cmd["params"]["expandAliases"] = kwargs["expandAliases"]

        return cmd

    async def post(self, jsonrpc, ofmt=None):
        res = await self.client.post("/command-api", json=jsonrpc)
        res.raise_for_status()
        body = res.json()

        commands = jsonrpc["params"]["cmds"]

        if ofmt == "text":

            def get_output(_cmd_r):
                return _cmd_r["output"]

        else:

            def get_output(_cmd_r):
                return _cmd_r

        if (err_data := body.get("error")) is None:
            return [
                CommandResults(
                    ok=True, command=commands[cmd_i], output=get_output(cmd_res)
                )
                for cmd_i, cmd_res in enumerate(body["result"])
            ]

        post_res = list()
        cmd_data = err_data["data"]
        len_data = len(cmd_data)
        err_at = len_data - 1
        err_msg = err_data["message"]

        # commands the passed
        for cmd_i, cmd in enumerate(commands[:err_at]):
            post_res.append(
                CommandResults(ok=True, command=cmd, output=get_output(cmd_data[cmd_i]))
            )

        # the command that failed
        post_res.append(
            CommandResults(ok=False, command=commands[err_at], output=err_msg)
        )

        # all other commands not executed
        for cmd in commands[err_at + 1 :]:
            post_res.append(CommandResults(ok=False, command=cmd, output=None))

        return post_res


class Device(object):
    def __init__(
        self,
        host: AnyStr,
        creds: Tuple[str, str],
        proto: Optional[AnyStr] = "https",
        port=None,
        private=None,
    ):
        self.host = host
        self.private = private
        self.api = Transport(host=host, creds=creds, proto=proto, port=port)

    async def exec(self, commands: List[AnyStr], **kwargs) -> List[CommandResults]:
        """
        Execute a list of operational commands and return the output as a list of CommandResults.
        """
        xcmd = self.api.form_command(commands=commands, **kwargs)
        return await self.api.post(xcmd, **kwargs)

    async def get_config(self, ofmt="text") -> List[CommandResults]:
        xmcd = self.api.form_command(commands=["show running-config"], ofmt=ofmt)
        return await self.api.post(xmcd, ofmt=ofmt)

    async def push_config(
        self, contents: str, enter_cmds=None, exit_cmds=None, ofmt="text"
    ) -> List[CommandResults]:
        config_cmds = contents.strip().splitlines()
        if not enter_cmds:
            config_cmds.insert(0, "configure")
        else:
            config_cmds[0:0] = enter_cmds

        if exit_cmds:
            config_cmds.extend(exit_cmds)

        jsonrpc = self.api.form_command(commands=config_cmds, ofmt=ofmt)
        return await self.api.post(jsonrpc=jsonrpc, ofmt=ofmt)



# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------
import asyncio
from typing import Optional, List, Union, Dict, AnyStr
from socket import getservbyname

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import httpx

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from .errors import EapiCommandError

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------


__all__ = ["Device"]


# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------


class Device(httpx.AsyncClient):
    """
    The Device class represents the async JSON-RPC client that communicates with
    an Arista EOS device.  This class inherits directly from the
    httpx.AsyncClient, so any initialization options can be passed directly.
    """

    auth = None
    EAPI_OFMT_OPTIONS = ("json", "text")
    EAPI_DEFAULT_OFMT = "json"

    def __init__(
        self,
        host: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        proto: Optional[str] = "https",
        port=None,
        **kwargs,
    ):
        """
        Initializes the Device class.  As a subclass to httpx.AsyncClient, the
        Caller can provide any of those initializers.  Specific paramertes for
        Device class are all optional and described below.

        Parameters
        ----------
        host: Optional[str]
            The EOS target device, either hostname (DNS) or ipaddress.

        username: Optional[str]
            The login user-name; requires the password parameter.

        password: Optional[str]
            The login password; requires the username parameter.

        proto: Optional[str]
            The protocol, http or https, to communicate eAPI with the device.

        port: Optional[Union[str,int]]
            If not provided, the proto value is used to lookup the associated
            port (http=80, https=443).  If provided, overrides the port used to
            communite with the device.

        Other Parameters
        ----------------
        base_url: str
            If provided, the complete URL to the device eAPI endpoint.

        auth:
            If provided, used as the httpx authorization initializer value. If
            not provided, then username+password is assumed by the Caller and
            used to create a BasicAuth instance.
        """

        self.port = port or getservbyname(proto)
        self.host = host
        kwargs.setdefault("base_url", httpx.URL(f"{proto}://{self.host}:{self.port}"))
        kwargs.setdefault("verify", False)

        if username and password:
            self.auth = httpx.BasicAuth(username, password)

        kwargs.setdefault("auth", self.auth)

        super(Device, self).__init__(**kwargs)
        self.headers["Content-Type"] = "application/json-rpc"

    async def check_connection(self) -> bool:
        """
        This function checks the target device to ensure that the eAPI port is
        open and accepting connections.  It is recommended that a Caller checks
        the connection before involing cli commands, but this step is not
        required.

        Returns
        -------
        True when the device eAPI is accessible, False otherwise.
        """
        try:
            await asyncio.open_connection(self.host, port=self.port)
        except OSError:
            return False
        return True

    async def cli(
        self,
        command: Optional[AnyStr] = None,
        commands: Optional[List[AnyStr]] = None,
        ofmt: Optional[str] = None,
        suppress_error: Optional[bool] = False,
        version: Optional[Union[int, str]] = "latest",
        **kwargs,
    ):
        """
        Execute one or more CLI commands.

        Parameters
        ----------
        command: str
            A single command to execute; results in a single output response

        commands: List[str]
            A list of commands to executes; results in a list of output responses

        ofmt: str
            Either 'json' or 'text'; indicates the output fromat for the CLI commands.

        suppress_error: Optional[bool] = False
            When not False, then if the execution of the command would-have
            raised an EapiCommandError, rather than raising this exception this
            routine will return the value None.

            For example, if the following command would have raised
            EapiCommandError, now response would be set to None instead.

                response = dev.cli(..., suppress_error=True)

        version: Optional[int | string]
            By default the eAPI will use "version 1" for all API object models.
            This driver will, by default, always set version to "latest" so
            that the behavior matches the CLI of the device.  The caller can
            override the "latest" behavior by explicity setting the version.


        Other Parameters
        ----------------
        autoComplete: Optional[bool] = False
            Enabled/disables the command auto-compelete feature of the EAPI.  Per the
            documentation:
                Allows users to use shorthand commands in eAPI calls. With this
                parameter included a user can send 'sh ver' via eAPI to get the
                output of 'show version'.

        expandAliases: Optional[bool] = False
            Enables/disables the command use of User defined alias.  Per the
            documentation:
                Allowed users to provide the expandAliases parameter to eAPI
                calls. This allows users to use aliased commands via the API.
                For example if an alias is configured as 'sv' for 'show version'
                then an API call with sv and the expandAliases parameter will
                return the output of show version.

        Returns
        -------
        One or List of output respones, per the description above.
        """
        if not any((command, commands)):
            raise RuntimeError("Required 'command' or 'commands'")

        jsonrpc = self.jsoncrpc_command(
            commands=[command] if command else commands,
            ofmt=ofmt,
            version=version,
            **kwargs,
        )

        try:
            res = await self.jsonrpc_exec(jsonrpc)
            return res[0] if command else res
        except EapiCommandError as eapi_error:
            if suppress_error:
                return None
            raise eapi_error

    def jsoncrpc_command(self, commands, ofmt, version, **kwargs) -> dict:
        """Used to create the JSON-RPC command dictionary object"""

        cmd = {
            "jsonrpc": "2.0",
            "method": "runCmds",
            "params": {
                "version": version,
                "cmds": commands,
                "format": ofmt or self.EAPI_DEFAULT_OFMT,
            },
            "id": str(kwargs.get("req_id") or id(self)),
        }
        if "autoComplete" in kwargs:
            cmd["params"]["autoComplete"] = kwargs["autoComplete"]

        if "expandAliases" in kwargs:
            cmd["params"]["expandAliases"] = kwargs["expandAliases"]

        return cmd

    async def jsonrpc_exec(self, jsonrpc: dict) -> List[Union[Dict, AnyStr]]:
        """
        Execute the JSON-RPC dictionary object.

        Parameters
        ----------
        jsonrpc: dict
            The JSON-RPC as created by the `meth`:jsonrpc_command().

        Raises
        ------
        EapiCommandError
            In the event that a command resulted in an error response.

        Returns
        -------
        The list of command results; either dict or text depending on the
        JSON-RPC format pameter.
        """
        res = await self.post("/command-api", json=jsonrpc)
        res.raise_for_status()
        body = res.json()

        commands = jsonrpc["params"]["cmds"]
        ofmt = jsonrpc["params"]["format"]

        get_output = (lambda _r: _r["output"]) if ofmt == "text" else (lambda _r: _r)

        # if there are no errors then return the list of command results.

        if (err_data := body.get("error")) is None:
            return [get_output(cmd_res) for cmd_res in body["result"]]

        # ----------------------------------------------------------------
        # if we are here, then there were some command errors.  Raise a
        # RuntimeError exception with args (commands that failed, passed,
        # not-executed).
        # ----------------------------------------------------------------

        cmd_data = err_data["data"]
        len_data = len(cmd_data)
        err_at = len_data - 1
        err_msg = err_data["message"]

        raise EapiCommandError(
            passed=[
                get_output(cmd_data[cmd_i])
                for cmd_i, cmd in enumerate(commands[:err_at])
            ],
            failed=commands[err_at],
            errmsg=err_msg,
            not_exec=commands[err_at + 1 :],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioeapi', 'asynceapi']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.3,<0.24.0']

setup_kwargs = {
    'name': 'aio-eapi',
    'version': '0.4.0',
    'description': 'Arista EOS API asyncio client',
    'long_description': '# Arista EOS API asyncio Client\n\nThis repository contains an Arista EOS asyncio client.\n\n### Quick Example\n\nThie following shows how to create a Device instance and run a list of\ncommands.\n\nDevice will use HTTPS transport by default.  The Device instance supports the\nfollowing initialization parameters:\n\n   * `host` - The device hostname or IP address\n   * `username` - The login username\n   * `password` - The login password\n   * `proto` - *(Optional)* Choose either "https" or "http", defaults to "https"\n   * `port` - *(Optional)* Chose the protocol port to override proto default\n\nThe Device class inherits directly from httpx.AsyncClient.  As such, the Caller\ncan provide any initialization parameters.  The above specific parameters are\nall optional.\n\n```python\nimport json\nfrom aioeapi import Device\n\nusername = \'dummy-user\'\npassword = \'dummy-password\'\n\nasync def run_test(host):\n    dev = Device(host=host, username=username, password=password)\n    res = await dev.cli(commands=[\'show hostname\', \'show version\'])\n    json.dumps(res)\n```\n',
    'author': 'Jeremy Schulman',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
